.bilitube iframe {
    width: 100%;
    aspect-ratio: 1.2877;
}

.hero {
    text-align: center;
}

.hero .action, .hero p, .hero img, .hero h1 {
    margin: 1.8rem auto;
}

.hero img {
    max-width: 100%;
    max-height: 280px;
    display: block;
    margin: 3rem auto 1.5rem;
    background: none!important;
}

.hero p {
    max-width: 35rem;
    font-size: 1.6rem;
    line-height: 1.3;
    color: #6a8bad;
}

.hero .action-button {
    display: inline-block;
    font-size: 1.2rem;
    color: #fff;
    background-color: #3eaf7c;
    padding: .8rem 1.6rem;
    border-radius: 4px;
    transition: background-color .1s ease;
    box-sizing: border-box;
    border-bottom: 1px solid #389d70;
}

.features {
    border-top: 1px solid #eaecef;
    padding: 1.2rem 0;
    margin-top: 2.5rem;
    display: flex;
    flex-wrap: wrap;
    align-items: flex-start;
    align-content: stretch;
    justify-content: space-between;
}

.features section {
    flex-grow: 1;
    flex-basis: 30%;
    max-width: 30%;
}

@media (max-width: 719px) {
    .features {
        flex-direction: column;
    }

    .features section {
        max-width: 100%;
        padding: 0 2.5rem;
    }
}
